/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2017 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://oss.oracle.com/licenses/CDDL+GPL-1.1
 * or LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
package org.glassfish.servlet.weather_station_war;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.servlet.ServletException;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.util.Enumeration;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileReader;
import java.io.Writer;
import java.io.Reader;
import java.io.StringWriter;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import org.apache.commons.io.IOUtils;
import java.util.Arrays;

/**
 * This Servlet acts as client to create HTTP connection to server, sends two
 * parts of data with 3 seconds sleeping in between, in order to simulate data
 * blocking.
 * @author Daniel Guo
 */
@WebServlet(name = "ClientTest", urlPatterns = {"/*"})
public class ClientTest extends HttpServlet {

    OutputStream output = null;
    InputStream input = null;


    public String convertStreamToString(InputStream is)
                throws IOException {
            /*
             * To convert the InputStream to String we use the
             * Reader.read(char[] buffer) method. We iterate until the
             * Reader return -1 which means there's no more data to
             * read. We use the StringWriter class to produce the string.
             */
            if (is != null) {
                Writer writer = new StringWriter();

                char[] buffer = new char[1024];
                try
                {
                    Reader reader = new BufferedReader(
                            new InputStreamReader(is, "UTF-8"));
                    int n;
                    while ((n = reader.read(buffer)) != -1) 
                    {
                        writer.write(buffer, 0, n);
                    }
                }
                finally 
                {
                    is.close();
                }
                return writer.toString();
            } else {       
                return "";
            }
        }
    
    /**
     * Processes requests for both HTTP
     * <code>GET</code> and
     * <code>POST</code> methods.
     *
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    protected void processRequest(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        response.setContentType("text/html;charset=UTF-8");

        PrintWriter out = response.getWriter();

        out.println("<html>");
        out.println("<head>");
	out.println("<link rel=\"stylesheet\" href=\"https://cdnjs.cloudflare.com/ajax/libs/mini.css/2.3.7/mini-default.min.css\">");
	out.println("<link rel=\"stylesheet\" href=\"https://maxcdn.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css\">");
	out.println("<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">");
        out.println("<title>weather-station-war</title>");
        out.println("</head>");
        out.println("<body>");

        String urlPath = "http://"
                + request.getServerName()
                + ":" + request.getLocalPort() //default http port is 8080
                + request.getContextPath()
                + "/ServerTest";

        URL url = new URL(urlPath);

        HttpURLConnection conn = (HttpURLConnection) url.openConnection();
        conn.setDoOutput(true);
        conn.setRequestMethod("POST");
        conn.setChunkedStreamingMode(2);
        conn.setRequestProperty("Content-Type", "text/plain");
        conn.connect();

        try {
            output = conn.getOutputStream();
            // Sending the first part of data to server
            String firstPart = "Hello";
            out.println("<p>Sending to server: " + firstPart + "</p>");
            out.flush();
            writeData(output, firstPart);

            Thread.sleep(2000);

            // Sending the second part of data to server
            String secondPart = "World";
            out.println("<p>Sending to server: " + secondPart + "</p></br>");
            out.flush();
            writeData(output, secondPart);

	    // Getting the echo data from server
            input = conn.getInputStream();
            out.print("<p>");
            printEchoData(out, input);
            out.print("</p>");

            out.println("<p>Please check server log for detail</p>");
            out.flush();

	    // Start parsing of parameters
	    Enumeration paramNames = request.getParameterNames();
            out.print("<h1> Parameter Analysis...</h1>");
            out.println("<table border=2 cellpadding=5 cellspacing=5>");                
            //out.println("<table>");
            out.println("<tr> <th>Parameter Name</th>" +
                        "<th>Parameter Value</th></tr>");

	    String city = "278";
	    
            while(paramNames.hasMoreElements())
            {
                    String paramName = (String)paramNames.nextElement();
                    out.print("<tr><td>" + paramName + "\n<td>");
                    String[] paramValues = request.getParameterValues(paramName);
                    if (paramValues.length == 1)
                    {
                            String paramValue = paramValues[0];
                            if (paramValue.length() == 0)
                                    out.println("No Value");
                            else {
                                    out.println(paramValue);
				    // Store the city code
				    if (paramName.equals("City"))
				        city = paramValue;
			    }
                    }
                    else
                    {
                            out.println("<ul>");
                            for(int i=0; i<paramValues.length; i++)
                            {
                                    out.println("<li>" + paramValues[i] + "</li>");
                            }
                            out.println("</ul>");
                    }
            }
            out.println("</table>");
	    out.flush();

	    // Query the Internet
            // City code 278 is New-york
            //String city = "278";

            ProcessBuilder pb = new ProcessBuilder(
                "/usr/bin/curl",
                "-s",
                "http://worldweather.wmo.int/fr/json/" + city + "_fr.xml");

            pb.directory(new File("/tmp/your_user_name/Cities"));
            pb.redirectErrorStream(true);
            Process p = pb.start();
            InputStream is = p.getInputStream();

            FileOutputStream outputStream = new FileOutputStream(
                "/tmp/your_user_name/Cities/" + city + "_fr.json");

            BufferedInputStream bis = new BufferedInputStream(is);
            byte[] bytes = new byte[100];
            int numberByteReaded;
            while ((numberByteReaded = bis.read(bytes, 0, 100)) != -1) {
                 outputStream.write(bytes, 0, numberByteReaded);
                 Arrays.fill(bytes, (byte) 0);
            }
	    outputStream.flush();
            outputStream.close();

	    //String s = new String(bytes);
	    //out.println("<p>Answer: "+s+" ---</p>");

	    out.println("<p>The City code is <b>"+city+"</b></p>");

	    //for(Byte i : bytes)
	    //	out.print(Byte.toString(i));
	    //out.println();
	    //out.println("<br><br>");
            out.println("<h2>Content of the file handling the temporary result</h2>");
	    String s = "";
	    try {
		FileReader reader = new FileReader("/tmp/your_user_name/Cities/" + city + "_fr.json");
		int character;
		out.print("<p>");
		while ((character = reader.read()) != -1) {
		    out.print((char) character);
		    s = s + ((char) character);
		}
                out.print("</p>");
		reader.close();

		out.println("<h2>We rebuild the String and we print it!</h2>");
		out.println("<p>Answer: "+s+"---</p>");

		//
		// Now we start the Python code to parse the answer
		// (we build a Python dictionary)
		//
		ProcessBuilder pbb = new ProcessBuilder(
		   "python",
		   "/Users/christophecerin/glassfish5/samples/servlet/weather-station-war/src/main/java/org/glassfish/servlet/weather_station_war/parse.py",
                   "/tmp/your_user_name/Cities/"+city+"_fr.json");
                pbb.redirectErrorStream(true);
                Process pp = pbb.start();
                InputStream iss = pp.getInputStream();
		String theString = IOUtils.toString(iss, "UTF-8");
		out.println("<h2>Here we do not use explicitely a temporary file</h2>");
		out.println("<p>"+theString+"</p>");
                out.println("<h2>A Potential shared ressource access problem</h2>");
                out.println("<p>In the case of using implicitly files for storing temporary results, it remains to protect the file names because in case of multiple queries on the same city we may have a concurrent access to the same file (name). We call it a race condition. Please, generate a random number to put a unique label for the files in order to distinguish them.</p>");

	    } catch (IOException e) {
		e.printStackTrace();
	    }

            out.flush();
   
        } catch (IOException ioException) {
            Logger.getLogger(ReadListenerImpl.class.getName()).log(Level.SEVERE,
                    "Please check the connection or url path", ioException);
        } catch (InterruptedException interruptedException) {
            Logger.getLogger(ReadListenerImpl.class.getName()).log(Level.SEVERE,
                    "Thread sleeping error", interruptedException);
        } finally {
            if (input != null) {
                try {
                    input.close();
                } catch (Exception ex) {
                }
            }
            if (output != null) {
                try {
                    output.close();
                } catch (Exception ex) {
                }
            }
        }

        out.println("</body>");
        out.println("</html>");

    }

    /**
     * Handles the HTTP
     * <code>GET</code> method.
     *
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        processRequest(request, response);
    }

    /**
     * Handles the HTTP
     * <code>POST</code> method.
     *
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
	        processRequest(request, response);
    }

    protected void writeData(OutputStream output, String data) throws IOException {
        if (data != null && !data.equals("") && output != null) {
            output.write(data.getBytes());
            output.flush();
        }
    }

    protected void printEchoData(PrintWriter out, InputStream input) throws IOException {
        while (input.available() > 0 && input != null && out != null) {
            out.print((char) input.read());
        }
        out.println("</br>");
    }
}
